# coding=utf-8
# Copyright (C) Duncan Macleod (2015)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""
#################################
How to write a configuration file
#################################

:mod:`gwdetchar.omega` can be used to process an arbitrary list of channels,
including primary gravitational wave strain channels and auxiliary sensors,
with arbitrary units and sample rates. Channels can be organized in contextual
blocks using an INI-formatted configuration file that must be passed at
runtime, which must include processing options for individual blocks. In a
given block, the following keywords are supported:

Keywords
--------

=======================  ======================================================
``name``                 The full name of this channel block, which will
                         appear as a section header on the output page
                         (optional)
``parent``               The `blockkey` of a section that the current block
                         should appear on the output page with (optional)
``q-range``              Range of quality factors (or Q) to search (required)
``frequency-range``      Range of frequencies to search (required)
``resample``             A sample rate (in Hz) to resample the input data to,
                         must be different from the original sample rate
                         (optional)
``frametype``            The type of frame files to read data from (will be
                         superceded by `source`, required if `source` is not
                         specified)
``source``               Path to a LAL-format cache pointing to frame files
``state-flag``           A data quality flag to require to be active before
                         processing this block (can be superceded by passing
                         ``--ignore-state-flags`` on the command-line;
                         optional)
``duration``             The duration of data to process in this block
                         (required)
``fftlength``            The FFT length to use in computing an ASD for
                         whitening with an overlap of `fftlength/2` (required)
``search``               Duration (seconds) of search time window (optional)
``max-mismatch``         The maximum mismatch in time-frequency tiles
                         (optional)
``snr-threshold``        Threshold on SNR for plotting eventgrams (optional)
``dt``                   Maximum acceptable time delay from the primary
                         channel (optional)
``always-plot``          Always analyze this block regardless of channel
                         significance (optional; will be superceded by
                         `state-flag` unless `--ignore-state-flags` is passed)
``plot-time-durations``  Time-axis durations of omega scan plots (required)
``channels``             Full list of channels which appear in this block
                         (required)
=======================  ======================================================

If cross-correlation will be implemented, the user will also need to specify
a block whose blockkey is ``primary`` that includes only one ``channel``
and options for ``f-low`` (a high-pass corner frequency) and
``matched-filter-length``. State flags are always ignored for the primary.

An example using many of the above options would look something like this:

.. code-block:: ini

   [primary]
   ; the primary channel, which will be used as a matched-filter
   f-low = 4.0
   resample = 4096
   frametype = L1_HOFT_C00
   duration = 64
   fftlength = 8
   matched-filter-length = 2
   channel = L1:GDS-CALIB_STRAIN

   [GW]
   ; name of this block, which contains h(t)
   name = Gravitational Wave Strain
   q-range = 3.3166,150.0
   frequency-range = 4.0,2048
   resample = 4096
   frametype = L1_HOFT_C00
   state-flag = L1:DMT-GRD_ISC_LOCK_NOMINAL:1
   duration = 64
   fftlength = 8
   max-mismatch = 0.2
   snr-threshold = 5
   always-plot = True
   plot-time-durations = 1,4,16
   channels = L1:GDS-CALIB_STRAIN

   [CAL]
   ; a sub-block of channels with different options, but which should appear
   ; together with the block `GW` on the output page
   parent = GW
   q-range = 3.3166,150
   frequency-range = 4.0,Inf
   resample = 4096
   frametype = L1_R
   state-flag = L1:DMT-GRD_ISC_LOCK_NOMINAL:1
   duration = 64
   fftlength = 8
   max-mismatch = 0.35
   snr-threshold = 5.5
   always-plot = True
   plot-time-durations = 1,4,16
   channels = L1:CAL-DELTAL_EXTERNAL_DQ

.. note::

   The `blockkey` will appear in the navbar to identify channel blocks on the
   output page, with a scrollable dropdown list of channels in that block for
   ease of navigation.

   The `primary` block will only be used to design a matched-filter. To process
   this channel during the omega scan, it must be included again in a
   subsequent block.

   If running on a LIGO Data Grid (LDG) computer cluster, the `~detchar`
   account houses default configurations organized by subsystem.
"""

import sys
import ast
import os.path
import configparser
import numpy

from gwpy.detector import Channel

from .. import const
from ..io.html import FancyPlot

if sys.version_info < (3, 7):
    from collections import OrderedDict

__author__ = 'Alex Urban <alexander.urban@ligo.org>'
__credits__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


# -- define parser ------------------------------------------------------------

class OmegaConfigParser(configparser.ConfigParser):
    """Custom configuration parser for :mod:`gwdetchar.omega`

    Parameters
    ----------
    ifo : `str`, optional
        prefix of the interferometer to use, defaults to `None`

    defaults : `dict`, optional
        dictionary of default values to pass to the parser, default: ``{}``

    **kwargs : `dict`, optional
        additional keyword arguments to pass to `ConfigParser`

    Methods
    -------
    read:
        parse a list of INI-format configuration files
    get_channel_blocks:
        retrieve an ordered dictionary of contextual channel blocks, as
        organized in the source configuration
    """
    def __init__(self, ifo=None, defaults=dict(), **kwargs):
        if ifo is not None:
            defaults.setdefault('IFO', ifo)
        configparser.ConfigParser.__init__(self, defaults=defaults, **kwargs)

    def read(self, filenames):
        readok = configparser.ConfigParser.read(self, filenames)
        for f in filenames:
            if f not in readok:
                raise IOError("Cannot read file %r" % f)
        return readok
    read.__doc__ = configparser.ConfigParser.read.__doc__

    def get_channel_blocks(self):
        """Retrieve an ordered dictionary of channel blocks

        These blocks are organized contextually by the user, since they are
        read in and preserved from the source configuration.
        """
        # retrieve an ordered dictionary of channel blocks
        if sys.version_info >= (3, 7):  # python 3.7+
            return {s: OmegaChannelList(s, **self[s]) for s in self.sections()}
        else:
            return OrderedDict([(s, OmegaChannelList(s, **dict(self.items(s))))
                                for s in self.sections()])


# -- utilities ----------------------------------------------------------------

def get_default_configuration(ifo, gpstime):
    """Retrieve a default configuration file stored locally

    Parameters
    ----------
    ifo : `str`
        interferometer ID string, e.g. `'L1'`

    gpstime : `float`
        time of analysis in GPS second format
    """
    # find epoch
    epoch = const.gps_epoch(gpstime, default=const.latest_epoch())
    # find and parse configuration file
    if ifo == 'Network':
        return [os.path.expanduser(
            '~detchar/etc/omega/{epoch}/Network.ini'.format(epoch=epoch))]
    else:
        return [os.path.expanduser(
            '~detchar/etc/omega/{epoch}/{obs}-{ifo}_R-selected.ini'.format(
                epoch=epoch, obs=ifo[0], ifo=ifo))]


def get_fancyplots(channel, plottype, duration, caption=None):
    """Construct FancyPlot objects for output HTML pages

    Parameters
    ----------
    channel : `str`
        the name of the channel

    plottype : `str`
        the type of plot, e.g. 'raw_timeseries'

    duration : `str`
        duration of the plot, in seconds

    caption : `str`, optional
        a caption to render in the fancybox
    """
    plotdir = 'plots'
    chan = channel.replace('-', '_').replace(':', '-')
    filename = '%s/%s-%s-%s.png' % (plotdir, chan, plottype, duration)
    if not caption:
        caption = os.path.basename(filename)
    return FancyPlot(filename, caption)


# -- channel list objects -----------------------------------------------------

class OmegaChannel(Channel):
    """Customized `Channel` object for omega scan analyses

    Parameters
    ----------
    channelname : `str`
        name of this channel, e.g. `L1:GDS-CALIB_STRAIN`

    section : `str`
        configuration section to which this channel belongs

    params : `dict`
        parameters set in a configuration file
    """
    def __init__(self, channelname, section, **params):
        self.name = channelname
        frametype = params.get('frametype', None)
        super(OmegaChannel, self).__init__(
            channelname, frametype=frametype)
        if section != 'primary':
            self.qrange = tuple(
                [float(s) for s in params.get('q-range').split(',')])
            self.frange = tuple(
                [float(s) for s in params.get('frequency-range').split(',')])
            self.mismatch = float(params.get('max-mismatch', 0.2))
            self.snrthresh = float(params.get('snr-threshold', 5.5))
            self.always_plot = ast.literal_eval(
                params.get('always-plot', 'False'))
            self.pranges = [int(t) for t in params.get('plot-time-durations',
                                                       None).split(',')]
            self.plots = {}
            for plottype in ['timeseries_raw', 'timeseries_highpassed',
                             'timeseries_whitened', 'qscan_highpassed',
                             'qscan_whitened', 'qscan_autoscaled',
                             'eventgram_highpassed', 'eventgram_whitened',
                             'eventgram_autoscaled']:
                self.plots[plottype] = [get_fancyplots(self.name, plottype, t)
                                        for t in self.pranges]
        self.section = section
        self.params = params.copy()

    def save_loudest_tile_features(self, qgram, correlate=None, gps=0, dt=0.1):
        """Store properties of the loudest time-frequency tile

        Parameters
        ----------
        channel : `OmegaChannel`
            `OmegaChannel` object corresponding to this data stream

        qgram : `~gwpy.signal.qtransform.QGram`
            output of a Q-transform on whitened input

        correlate : `~gwpy.timeseries.TimeSeries`, optional
            output of a single phase matched-filter, or `None` if not requested

        gps : `float`, optional
            GPS time (seconds) of suspected transient, used only if `correlate`
            is not `None`, default: 0

        dt : `float`, optional
            maximum acceptable time delay (seconds) between the primary and
            `self`, used only if `correlate` is not `None`, default: 0.1

        Notes
        -----
        Attributes stored in-place include `Q`, `energy`, `snr`, `t`, and `f`,
        all corresponding to the loudest time-frequency tile contained in
        `qgram`.

        If `correlate` is not `None` then the maximum correlation amplitude,
        relative time delay, and standard deviation are stored as attributes
        `corr`, `delay`, and `stdev`, respectively.
        """
        # save parameters
        self.Q = numpy.around(qgram.plane.q, 1)
        self.energy = numpy.around(qgram.peak['energy'], 1)
        self.snr = numpy.around(qgram.peak['snr'], 1)
        self.t = numpy.around(qgram.peak['time'], 3)
        self.f = numpy.around(qgram.peak['frequency'], 1)
        if correlate is not None:
            stdev = correlate.std().value
            corr = correlate.abs().crop(gps - dt, gps + dt)
            self.corr = numpy.around(corr.max().value, 1)
            self.stdev = stdev  # used to reject high glitch rates
            delay = (corr.t0 + corr.argmax() * corr.dt).value - gps
            self.delay = int(delay * 1000)  # convert to ms

    def load_loudest_tile_features(self, table, correlated=False):
        """Load properties of the loudest time-frequency tile from a table

        Parameters
        ----------
        table : `~gwpy.table.Table`
            table of properties of the loudest time-frequency tile

        correlated : `bool`, optional
            boolean switch to determine if cross-correlation properties are
            included, default: `False`

        Notes
        -----
        Attributes stored in-place include `Q`, `energy`, `snr`, `t`, and `f`,
        all corresponding to the columns contained in `table`.

        If `correlated` is not `None` then the maximum correlation amplitude,
        relative time delay, and standard deviation are stored as attributes
        `corr`, `delay`, and `stdev`, respectively.
        """
        # save parameters
        self.Q = table['Q']
        self.energy = table['Energy']
        self.snr = table['SNR']
        self.t = table['Central Time']
        self.f = table['Central Frequency (Hz)']
        if correlated:
            self.corr = table['Correlation']
            self.stdev = table['Standard Deviation']
            self.delay = table['Delay (ms)']


class OmegaChannelList(object):
    """A conceptual list of `OmegaChannel` objects with common signal
    processing settings

    Parameters
    ----------
    key : `str`
        the unique identifier for this list, e.g. `'CAL'` for calibration
        channels

    params : `dict`
        parameters set in a configuration file
    """
    def __init__(self, key, **params):
        self.key = key
        self.parent = params.get('parent', None)
        self.name = params.get('name', None)
        self.duration = int(params.get('duration', 32))
        self.fftlength = int(params.get('fftlength', 2))
        self.resample = int(params.get('resample', 0))
        self.source = params.get('source', None)
        self.frametype = params.get('frametype', None)
        section = self.parent if self.parent else self.key
        if key == 'primary':
            self.length = float(params.get('matched-filter-length'))
            self.flow = float(params.get('f-low', 4))
            channelname = params.get('channel').strip()
            self.channel = OmegaChannel(channelname, section, **params)
        else:
            self.flag = params.get('state-flag', None)
            self.search = float(params.get('search', 0.5))
            self.dt = float(params.get('dt', 0.1))
            chans = params.get('channels', None).strip().split('\n')
            self.channels = [OmegaChannel(c, section, **params) for c in chans]
        self.params = params.copy()



# coding=utf-8
# Copyright (C) Duncan Macleod (2015)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for analysing optical scattering
"""

import numpy

from scipy.signal import savgol_filter

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = ('Siddharth Soni <siddharth.soni@ligo.org>, '
               'Alex Urban <alexander.urban@ligo.org>')

OPTIC_MOTION_CHANNELS = {
    'BS': ['SUS-BS_M1_DAMP_L_IN1_DQ',
           'SUS-BS_M2_WIT_L_DQ'],
    'ETMX': ['SUS-ETMX_M0_DAMP_L_IN1_DQ',
             'SUS-ETMX_R0_DAMP_L_IN1_DQ',
             'SUS-ETMX_L2_WIT_L_DQ'],
    'ETMY': ['SUS-ETMY_M0_DAMP_L_IN1_DQ',
             'SUS-ETMY_R0_DAMP_L_IN1_DQ',
             'SUS-ETMY_L2_WIT_L_DQ'],
    'IM1': ['SUS-IM1_M1_DAMP_L_IN1_DQ'],
    'IM2': ['SUS-IM2_M1_DAMP_L_IN1_DQ'],
    'IM3': ['SUS-IM3_M1_DAMP_L_IN1_DQ'],
    'IM4': ['SUS-IM4_M1_DAMP_L_IN1_DQ'],
    'ITMX': ['SUS-ITMX_M0_DAMP_L_IN1_DQ',
             'SUS-ITMX_R0_DAMP_L_IN1_DQ',
             'SUS-ITMX_L2_WIT_L_DQ'],
    'ITMY': ['SUS-ITMY_M0_DAMP_L_IN1_DQ',
             'SUS-ITMY_R0_DAMP_L_IN1_DQ',
             'SUS-ITMY_L1_WIT_L_DQ'],
    'MC1': ['SUS-MC1_M1_DAMP_L_IN1_DQ',
            'SUS-MC1_M3_WIT_L_DQ'],
    'MC2': ['SUS-MC2_M1_DAMP_L_IN1_DQ',
            'SUS-MC2_M3_WIT_L_DQ'],
    'MC3': ['SUS-MC3_M1_DAMP_L_IN1_DQ',
            'SUS-MC3_M3_WIT_L_DQ'],
    'OM1': ['SUS-OM1_M1_DAMP_L_IN1_DQ'],
    'OM2': ['SUS-OM2_M1_DAMP_L_IN1_DQ'],
    'OM3': ['SUS-OM3_M1_DAMP_L_IN1_DQ'],
    'OMC': ['SUS-OMC_M1_DAMP_L_IN1_DQ'],
    'PR2': ['SUS-PR2_M1_DAMP_L_IN1_DQ',
            'SUS-PR2_M3_WIT_L_DQ'],
    'PR3': ['SUS-PR3_M1_DAMP_L_IN1_DQ',
            'SUS-PR3_M3_WIT_L_DQ'],
    'PRM': ['SUS-PRM_M1_DAMP_L_IN1_DQ',
            'SUS-PRM_M3_WIT_L_DQ'],
    'RM1': ['SUS-RM1_M1_DAMP_L_IN1_DQ'],
    'RM2': ['SUS-RM2_M1_DAMP_L_IN1_DQ'],
    'ZM1': ['SUS-ZM1_M1_DAMP_L_IN1_DQ'],
    'ZM2': ['SUS-ZM2_M1_DAMP_L_IN1_DQ'],
    'OPO': ['SUS-OPO_M1_DAMP_L_IN1_DQ'],
    'OFI': ['SUS-OFI_M1_DAMP_L_IN1_DQ',
            'SUS-OFI_M1_DAMP_T_IN1_DQ'],
    'SR2': ['SUS-SR2_M1_DAMP_L_IN1_DQ',
            'SUS-SR2_M3_WIT_L_DQ'],
    'SR3': ['SUS-SR3_M1_DAMP_L_IN1_DQ',
            'SUS-SR3_M3_WIT_L_DQ'],
    'SRM': ['SUS-SRM_M1_DAMP_L_IN1_DQ',
            'SUS-SRM_M3_WIT_L_DQ'],
    'TMSX': ['SUS-TMSX_M1_DAMP_L_IN1_DQ'],
    'TMSY': ['SUS-TMSY_M1_DAMP_L_IN1_DQ'],
}

TRANSMON_CHANNELS = ['ASC-X_TR_B_NSUM_OUT_DQ',
                     'ASC-Y_TR_B_NSUM_OUT_DQ']

FREQUENCY_MULTIPLIERS = range(1, 5)


def get_fringe_frequency(series, multiplier=2.0):
    """Predict scattering fringe frequency from the derivative of a timeseries

    Parameters
    ----------
    series : `~gwpy.timeseries.TimeSeries`
        timeseries record of relative motion

    multiplier : `float`
        harmonic number of fringe frequency

    Returns
    -------
    fringef : `~gwpy.timeseries.TimeSeries`
        timeseries record of fringe frequency

    See Also
    --------
    scipy.signal.savgol_filter
        for an implementation of the Savitzky-Golay filter
    """
    velocity = type(series)(savgol_filter(series.value, 5, 2, deriv=1))
    velocity.__array_finalize__(series)
    fringef = numpy.abs(multiplier * 2. / 1.064 * velocity *
                        velocity.sample_rate.value)
    fringef.override_unit('Hz')
    return fringef


def get_blrms(series, flow=4.0, fhigh=10.0, stride=1, whiten=True,
              fftlength=4, overlap=2, **kwargs):
    """Compute the whitened, band-limited RMS of a `TimeSeries`

    Parameters
    ----------
    series  : `~gwpy.timeseries.TimeSeries`
        the input `TimeSeries` data

    flow : `float`, optional
        lower limit (Hz) of the passband, default: 4.0

    fhigh : `float`, optional
        upper limit (Hz) of the passband, default: 10.0

    stride: `float`, optional
        RMS integration length (seconds), default: 1

    whiten : `bool`, optional
        boolean switch to enable (`True`) or disable (`False`)
        whitening of the input, default: `True`

    fftlength : `float`, optional
        FFT integration length (seconds), default: 4

    overlap : `float`, optional
        FFT overlap length (seconds), default: 2

    **kwargs : `dict`, optional
        additional keyword arguments to `TimeSeries.whiten`

    Returns
    -------
    wblrms : `~gwpy.timeseries.TimeSeries`
        whitened, band-limited RMS trends of the input `TimeSeries`

    See Also
    --------
    gwpy.timeseries.TimeSeries.whiten
        for the underlying whitening scheme
    gwpy.timeseries.TimeSeries.rms
        for the underlying root-mean-square (RMS) estimation method
    """
    if whiten:
        series = series.whiten(fftlength=fftlength, overlap=overlap, **kwargs)
    bpseries = series.bandpass(flow, fhigh)
    return bpseries.rms(stride)


def get_segments(series, threshold, name=None, pad=0):
    """Generate data-quality segments by thresholding a `TimeSeries`

    Parameters
    ----------
    series  : `~gwpy.timeseries.TimeSeries`
        the input `TimeSeries` data

    threshold : `float`
        the threshold value for active data-quality segments

    name : `str`, optional
        name of the data-quality flag, defaults to `series.name`

    pad : `float`, optional
        length (seconds) by which to pad active segments, default: 0

    Returns
    --------
    threshflag : `~gwpy.segments.DataQualityFlag`
        the populated data-quality flag
    """
    if series.value.max() < threshold:
        from gwpy.segments import DataQualityFlag
        return DataQualityFlag(name, known=[series.span])
    else:
        thresh = series >= threshold * series.unit
        threshflag = thresh.to_dqflag(name or series.name)
        threshflag.protract(pad)
        return threshflag.coalesce()



# -*- coding: utf-8 -*-
# Copyright (C) Alex Urban (2019)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gwdetchar.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwdetchar.omega.html`
"""

import os
import shutil
from io import StringIO

from .. import (config, html)
from ..._version import get_versions
from ...utils import parse_html

__author__ = 'Alex Urban <alexander.urban@ligo.org>'
__credits__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


# global test objects

VERSION = get_versions()['version']
COMMIT = get_versions()['full-revisionid']

HTML_HEADER = """<nav class="navbar fixed-top navbar-expand-md navbar-{ifo} shadow-sm">
<div class="container-fluid">
<div class="navbar-brand border border-white rounded">{IFO} &Omega;-scan</div>
<button class="navbar-toggler navbar-toggler-right" type="button" data-toggle="collapse" data-target=".navbar-collapse">
<span class="navbar-toggler-icon"></span>
</button>
<div class="collapse navbar-collapse justify-content-between">
<ul class="nav navbar-nav mr-auto">
<li class="nav-item navbar-text">
0
</li>
<li class="nav-item">
<a href="#" class="nav-link">Summary</a>
</li>
<li class="nav-item dropdown">
<a href="#" class="nav-link dropdown-toggle" role="button" data-toggle="dropdown">GW</a>
<div class="dropdown-menu dropdown-1-col shadow">
<div class="row">
<div class="col-sm-12 col-md-12">
<h6 class="dropdown-header">Gravitational-Wave Strain</h6>
<a href="#x1-test-aux" class="dropdown-item">X1:TEST-AUX</a>
</div>
</div>
</div>
</li>
</ul>
<ul class="nav navbar-nav">
<li class="nav-item dropdown">
<a class="nav-link dropdown-toggle" href="#" role="button" data-toggle="dropdown">Links</a>
<div class="dropdown-menu dropdown-menu-right shadow">
<h6 class="dropdown-header">Internal</h6>
<a href="about" class="dropdown-item">About this page</a>
<div class="dropdown-divider"></div>
<h6 class="dropdown-header">External</h6>
<a href="https://ldas-jobs.ligo-la.caltech.edu/~detchar/summary/day/19800106" class="dropdown-item" target="_blank">LLO Summary Pages</a>
<a href="https://alog.ligo-la.caltech.edu/aLOG" class="dropdown-item" target="_blank">LLO Logbook</a>
</div>
</li>
</ul>
</div>
</div>
</nav>"""  # noqa: E501

CONFIGURATION = u"""
[primary]
; the primary channel, which will be used as a matched-filter
f-low = 4.0
resample = 4096
frametype = X1_HOFT
duration = 64
fftlength = 8
matched-filter-length = 6
channel = X1:TEST-STRAIN

[GW]
; name of this block, which contains h(t)
name = Gravitational-Wave Strain
q-range = 3.3166,150.0
frequency-range = 4.0,1024
resample = 4096
frametype = X1_HOFT
state-flag = X1:OBSERVING:1
duration = 64
fftlength = 8
max-mismatch = 0.2
snr-threshold = 5
always-plot = True
plot-time-durations = 4
channels = X1:TEST-AUX
"""

CONFIG_FILE = StringIO(CONFIGURATION)
CP = config.OmegaConfigParser(ifo='X1')
CP.read_file(CONFIG_FILE)
BLOCKS = CP.get_channel_blocks()
PRIMARY = BLOCKS['primary']
GW = BLOCKS['GW']

# set analyzed channel list
ANALYZED = {'GW': {
    'name': 'Gravitational-Wave Strain',
    'channels': GW.channels,
}}
for channel in ANALYZED['GW']['channels']:
    channel.Q = 5
    channel.t = 0
    channel.f = 100
    channel.energy = 1000
    channel.snr = 44.7
    channel.corr = 100
    channel.stdev = 1
    channel.delay = 0

BLOCK_HTML = """<div class="card card-x1 mb-5 shadow-sm">
<div class="card-header pb-0">
<h5 class="card-title">GW: Gravitational-Wave Strain</h5>
</div>
<div class="card-body">
<div class="list-group">
<div class="list-group-item flex-column align-items-start">
<h5 class="card-title" id="x1-test-aux"><a class="cis-link" href="https://cis.ligo.org/channel/byname/X1:TEST-AUX" title="CIS entry for X1:TEST-AUX" target="_blank">X1:TEST-AUX</a></h5>
<div class="row">
<div class="col-sm-12 col-md-7">
<table class="table table-sm table-hover table-bordered  d-none d-lg-table">
<thead>
<tr>
<th scope="col">GPS Time</th>
<th scope="col">Frequency</th>
<th scope="col">Q</th>
<th scope="col">Energy</th>
<th scope="col">SNR</th>
<th scope="col">Correlation</th>
<th scope="col">Delay</th>
</tr>
</thead>
<tbody>
<tr>
<td>0</td>
<td>100 Hz</td>
<td>5</td>
<td>1000</td>
<td>44.7</td>
<td>100</td>
<td>0 ms</td>
</tr>
</tbody>
</table>
</div>
<div class="col-sm-12 col-md-5">
<div class="btn-group flex-wrap" role="group">
<div class="btn-group" role="group">
<button id="btnGroupTimeseries0" type="button" class="btn btn-x1 dropdown-toggle" data-toggle="dropdown">Timeseries</button>
<div class="dropdown-menu shadow" aria-labelledby="btnGroupTimeseries0">
<a class="dropdown-item image-switch" data-captions="[&quot;X1-TEST_AUX-timeseries_raw-4.png&quot;]" data-channel-name="X1-TEST_AUX" data-image-dir="plots" data-image-type="timeseries_raw" data-t-ranges="[&quot;4&quot;]">raw</a>
<a class="dropdown-item image-switch" data-captions="[&quot;X1-TEST_AUX-timeseries_highpassed-4.png&quot;]" data-channel-name="X1-TEST_AUX" data-image-dir="plots" data-image-type="timeseries_highpassed" data-t-ranges="[&quot;4&quot;]">highpassed</a>
<a class="dropdown-item image-switch" data-captions="[&quot;X1-TEST_AUX-timeseries_whitened-4.png&quot;]" data-channel-name="X1-TEST_AUX" data-image-dir="plots" data-image-type="timeseries_whitened" data-t-ranges="[&quot;4&quot;]">whitened</a>
</div>
</div>
<div class="btn-group" role="group">
<button id="btnGroupQscan0" type="button" class="btn btn-x1 dropdown-toggle" data-toggle="dropdown">Spectrogram</button>
<div class="dropdown-menu shadow" aria-labelledby="btnGroupQscan0">
<a class="dropdown-item image-switch" data-captions="[&quot;X1-TEST_AUX-qscan_highpassed-4.png&quot;]" data-channel-name="X1-TEST_AUX" data-image-dir="plots" data-image-type="qscan_highpassed" data-t-ranges="[&quot;4&quot;]">highpassed</a>
<a class="dropdown-item image-switch" data-captions="[&quot;X1-TEST_AUX-qscan_whitened-4.png&quot;]" data-channel-name="X1-TEST_AUX" data-image-dir="plots" data-image-type="qscan_whitened" data-t-ranges="[&quot;4&quot;]">whitened</a>
<a class="dropdown-item image-switch" data-captions="[&quot;X1-TEST_AUX-qscan_autoscaled-4.png&quot;]" data-channel-name="X1-TEST_AUX" data-image-dir="plots" data-image-type="qscan_autoscaled" data-t-ranges="[&quot;4&quot;]">autoscaled</a>
</div>
</div>
<div class="btn-group" role="group">
<button id="btnGroupEventgram0" type="button" class="btn btn-x1 dropdown-toggle" data-toggle="dropdown">Eventgram</button>
<div class="dropdown-menu shadow" aria-labelledby="btnGroupEventgram0">
<a class="dropdown-item image-switch" data-captions="[&quot;X1-TEST_AUX-eventgram_highpassed-4.png&quot;]" data-channel-name="X1-TEST_AUX" data-image-dir="plots" data-image-type="eventgram_highpassed" data-t-ranges="[&quot;4&quot;]">highpassed</a>
<a class="dropdown-item image-switch" data-captions="[&quot;X1-TEST_AUX-eventgram_whitened-4.png&quot;]" data-channel-name="X1-TEST_AUX" data-image-dir="plots" data-image-type="eventgram_whitened" data-t-ranges="[&quot;4&quot;]">whitened</a>
<a class="dropdown-item image-switch" data-captions="[&quot;X1-TEST_AUX-eventgram_autoscaled-4.png&quot;]" data-channel-name="X1-TEST_AUX" data-image-dir="plots" data-image-type="eventgram_autoscaled" data-t-ranges="[&quot;4&quot;]">autoscaled</a>
</div>
</div>
</div>
</div>
</div>
<div class="row scaffold">
<div class="col-sm-12">
<a href="plots/X1-TEST_AUX-qscan_whitened-4.png" id="a_X1-TEST_AUX_4" title="X1-TEST_AUX-qscan_whitened-4.png" class="fancybox" target="_blank" data-caption="X1-TEST_AUX-qscan_whitened-4.png" data-fancybox="gallery" data-fancybox-group="images">
<img id="img_X1-TEST_AUX_4" alt="X1-TEST_AUX-qscan_whitened-4.png" class="img-fluid w-100 lazy" data-src="plots/X1-TEST_AUX-qscan_whitened-4.png" />
</a>
</div>
</div>
</div>
</div>
</div>
</div>"""  # noqa: E501


# -- unit tests ---------------------------------------------------------------

def test_update_toc():
    toc = html.update_toc(dict(), GW.channels[0], GW.name)
    assert toc.keys() == ANALYZED.keys()
    assert toc['GW'].keys() == ANALYZED['GW'].keys()
    assert len(toc['GW']['channels']) == len(ANALYZED['GW']['channels'])
    assert toc['GW']['channels'][0].name == ANALYZED['GW']['channels'][0].name
    assert toc['GW']['name'] == ANALYZED['GW']['name']


def test_navbar():
    page = html.navbar('L1', 0, toc=ANALYZED)
    assert parse_html(str(page)) == parse_html(HTML_HEADER.format(
        ifo='l1', IFO='L1'))


def test_toggle_link():
    h1 = parse_html(html.toggle_link('timeseries_raw', GW.channels[0], [4]))
    h2 = parse_html(
        '<a class="dropdown-item image-switch" data-captions='
        '"[&quot;X1-TEST_AUX-timeseries_raw-4.png&quot;]" '
        'data-channel-name="X1-TEST_AUX" data-image-dir="plots" '
        'data-image-type="timeseries_raw" data-t-ranges="[&quot;4&quot;]">'
        'raw</a>'
    )
    assert h1 == h2


def test_write_summary_table(tmpdir):
    tmpdir.mkdir('data')
    wdir = str(tmpdir)
    os.chdir(wdir)
    html.write_summary_table(ANALYZED, correlated=True)
    shutil.rmtree(wdir)


def test_write_summary():
    page = html.write_summary('L1', 0, incomplete=True)
    assert parse_html(str(page)) == parse_html(
        '<div class="banner">\n<h2>Summary</h2>\n</div>\n<div class="row">'
        '\n<div class="col-md-5">\n<table class="table table-sm table-hover">'
        '\n<tbody>\n<tr>\n<th scope="row">Interferometer</th>'
        '\n<td>LIGO Livingston (L1)</td>\n</tr>\n<tr>'
        '\n<th scope="row">UTC Time</th>\n<td>1980-01-06 00:00:00</td>'
        '\n</tr>\n</tbody>\n</table>\n</div>\n<div class="col-sm-12 col-md-7">'
        '\n<div class="dropdown float-right d-none d-lg-block">'
        '\n<button type="button" class="btn btn-l1 dropdown-toggle" '
        'data-toggle="dropdown" aria-expanded="false" aria-haspopup="true">'
        'Download summary</button>\n<div class="dropdown-menu dropdown-menu-'
        'right shadow">\n<a href="data/summary.txt" download="L1_0_summary.txt'
        '" class="dropdown-item">txt</a>\n<a href="data/summary.csv" download='
        '"L1_0_summary.csv" class="dropdown-item">csv</a>'
        '\n<a href="data/summary.tex" download="L1_0_summary.tex" '
        'class="dropdown-item">tex</a>\n</div>\n</div>\n</div>\n</div>'
        '\n<div class="alert alert-l1 alert-dismissible fade show text-justify'
        ' shadow-sm">\n<button type="button" class="close" data-dismiss='
        '"alert"aria-label="Close">\n<span aria-hidden="true">&times;</span>\n'
        '</button>\n<strong>Note</strong>: This scan is in progress, and will '
        'auto-refresh every 60 seconds until completion.\n</div>')


def test_write_ranking():
    page = html.write_ranking(ANALYZED, PRIMARY.channel.name)
    h1 = parse_html(str(page))
    h2 = parse_html(
        '<div class="row">\n<div class="col-md-12">\n<p>Below are the top 5 '
        'channels ranked by matched-filter correlation within 100 ms of <a '
        'href="plots/primary.png" title="Whitened timeseries of the primary '
        'channel, X1:TEST-STRAIN." class="fancybox cis-link" target="_blank"'
        'data-fancybox="gallery" data-fancybox-group="images">X1:TEST-STRAIN'
        '</a>.</p>\n<table class="table table-sm table-hover table-bordered"'
        '>\n<thead>\n<tr>\n<th scope="col">Channel</th>\n<th scope="col">GPS'
        ' Time</th>\n<th scope="col">Frequency</th>\n<th scope="col">Q</th>'
        '\n<th scope="col">Energy</th>\n<th scope="col">SNR</th>'
        '\n<th scope="col">Correlation</th>\n<th scope="col">Delay</th>'
        '\n</tr>\n</thead>\n<tbody>\n<tr>\n<td><a title="X1:TEST-AUX" '
        'class="cis-link" href="#x1-test-aux">X1:TEST-AUX</a></td>'
        '\n<td>0</td>\n<td>100.0 Hz</td>\n<td>5</td>\n<td>1000</td>'
        '\n<td>44.7</td>\n<td>100</td>\n<td>0 ms</td>\n</tr>\n</tbody>'
        '\n</table>\n</div>\n</div>'
    )
    assert h1 == h2


def test_write_block():
    page = html.write_block('GW', ANALYZED['GW'], 'x1')
    assert parse_html(str(page)) == parse_html(BLOCK_HTML)


# -- end-to-end tests ---------------------------------------------------------

def test_write_qscan_page(tmpdir):
    tmpdir.mkdir('about')
    tmpdir.mkdir('data')
    tmpdir.mkdir('plots')
    base = str(tmpdir)
    config = os.path.join(base, 'config.ini')
    with open(config, 'w') as fobj:
        fobj.write(CONFIGURATION)
    os.chdir(base)
    htmlv = {
        'title': 'test',
        'refresh': True,
        'config': [config],
    }
    html.write_qscan_page('L1', 0, ANALYZED, **htmlv)
    html.write_qscan_page('L1', 0, ANALYZED, correlated=False, **htmlv)
    shutil.rmtree(base, ignore_errors=True)


def test_write_null_page(tmpdir):
    os.chdir(str(tmpdir))
    html.write_null_page('L1', 0, 'test')
    shutil.rmtree(str(tmpdir), ignore_errors=True)


def test_write_about_page(tmpdir):
    base = str(tmpdir)
    config = os.path.join(base, 'config.ini')
    with open(config, 'w') as fobj:
        fobj.write(CONFIGURATION)
    os.chdir(base)
    html.write_about_page('L1', 0, [config], outdir='about')
    shutil.rmtree(base, ignore_errors=True)



# coding=utf-8
# Copyright (C) Duncan Macleod (2015)
#
# This file is part of the GW DetChar python package.
#
# GW DetChar is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GW DetChar is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GW DetChar.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for accessing information from CDS
"""

from urllib import request

RTDCUID_URL = (
    'https://daqsvn.ligo-la.caltech.edu/websvn/filedetails.php?'
    'repname=daq_maps&path=%2F{ifo}%2Frtdcuid'
)

ADCLIST_URL = (
    'https://daqsvn.ligo-la.caltech.edu/websvn/filedetails.php?'
    'repname=daq_maps&path=%2F{ifo}%2Fadclists%2F{model}_adclist.txt'
)

DCUID_MAP = {}
ADC_MAP = {}

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'


def get_dcuid_map(ifo):
    import bs4
    html = request.urlopen(RTDCUID_URL.format(ifo=ifo.lower())).read()
    soup = bs4.BeautifulSoup(html, 'lxml')
    listing = soup.find('div', attrs={'class': 'listing'})
    mapping = {}
    for code in listing.find_all('code'):
        try:
            fec, model = str(code.get_text()).rstrip(' \n').split()
            fec = int(fec)
        except (ValueError, TypeError):
            continue
        else:
            mapping[fec] = model
    return mapping


def get_adclist(ifo, model):
    import bs4
    html = request.urlopen(ADCLIST_URL.format(ifo=ifo.lower(),
                                              model=model.lower())).read()
    soup = bs4.BeautifulSoup(html, 'lxml')
    listing = soup.find('div', attrs={'class': 'listing'})
    mapping = {}
    for code in listing.find_all('code'):
        try:
            card, slot, channel, part = str(
                code.get_text()).rstrip(' \n').split()
            card = int(card)
            slot = int(slot)
        except (ValueError, TypeError):
            continue
        else:
            mapping[(card, slot)] = channel
    return mapping


def model_name_from_dcuid(ifo, dcuid):
    global DCUID_MAP
    if ifo not in DCUID_MAP:
        DCUID_MAP[ifo] = get_dcuid_map(ifo)
    try:
        return DCUID_MAP[ifo][dcuid]
    except KeyError as e:
        e.args = ('No model name associated with DCUID=%d' % dcuid,)
        raise


def dcuid_from_model_name(ifo, model):
    global DCUID_MAP
    if ifo not in DCUID_MAP:
        DCUID_MAP[ifo] = get_dcuid_map(ifo)
    for dcuid, name in DCUID_MAP[ifo].items():
        if model.lower() == name:
            return dcuid
    raise KeyError("No DCUID associated with model=%r" % model)


def get_adc_channel(ifo, model, card, slot):
    global ADC_MAP
    ADC_MAP.setdefault(ifo, {})
    if model not in ADC_MAP[ifo]:
        ADC_MAP[ifo][model] = get_adclist(ifo, model)
    try:
        return ADC_MAP[ifo][model][(card, slot)]
    except KeyError as e:
        e.args = ('No channel associated with card %d, slot %d for %s'
                  % (card, slot, model),)
        raise


def get_real_channel(adcchannel):
    main = str(adcchannel).split('-', 1)[1]
    ifo = str(adcchannel).split(':')[0]
    dcuid, type_, _, _, card, slot = main.split('_')
    if type_ != 'ADC':
        raise ValueError("No 'real' channel map for non-ADC channels")
    modelname = model_name_from_dcuid(ifo, int(dcuid))
    return get_adc_channel(ifo, modelname, int(card), int(slot))
